﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using IdeaBlade.EntityModel;
using DomainModel;
using System.Collections;
using IdeaBlade.Linq;
using System.Linq.Expressions;

namespace Console01 {

  /// <summary>
  /// Code Snippets for topic document "Business Object Persistence"
  /// </summary>
  class Program {
    /// <summary>
    /// 
    /// </summary>
    /// <param name="args"></param>
    /// <remarks>
    ///  We suggest that you comment out all method calls in method Main, then uncomment
    ///  them one-by-one for examination and exercise.
    /// </remarks>
    static void Main(string[] args) {
      Program aProgram = new Program();

      ////---------------------------------
      ////Basic Queries
      ////---------------------------------
      //aProgram.BasicQuerySyntaxQuery();
      //aProgram.BasicMethodSyntaxQuery();
      //aProgram.MethodSyntaxShortForm();
      //aProgram.MixedQueryAndMethodSyntax();
      //aProgram.QueryWithExplicitlyTypedReturnValue();
      //aProgram.RetrieveFirstCustomer();

      ////---------------------------------
      //// ***** Forcing Execution
      ////---------------------------------
      //aProgram.ForceImmediateExecution();
      //aProgram.ForceRetrievalUsingForEach();

      ////---------------------------------
      ////Skip and Take
      ////---------------------------------
      //aProgram.QueryWithSkipAndTake();

      ////---------------------------------
      ////Queries with WITH
      ////---------------------------------
      //aProgram.QueriesWithWITH();

      ////---------------------------------
      ////Using ToQuery()
      ////---------------------------------
      //aProgram.UsingToQueryPt01();
      //aProgram.UsingToQueryPt02();

      ////---------------------------------
      ////ESQL and Stored Proc Queries
      ////---------------------------------
      //aProgram.PassThruEsqlQuery();
      // aProgram.StoredProcQuery(); - stored procs not yet supported.
      //aProgram.EsqlBasic();
      //aProgram.EsqlWithParameter();

      ////---------------------------------
      ////PredicateBuilder 
      ////---------------------------------
      //aProgram.ProductsWithNamesThatContainSpecifiedString();
      //aProgram.ProductsWithNamesThatContainSpecifiedStrings();
      //aProgram.PredicateBuilder01();
      //aProgram.PredicateBuilderMiscExamples();
      //aProgram.PredicateDescriptions01();
      //aProgram.CompositePredicateDescription();
      //aProgram.CompositePredicateDescriptionToLambda();
      //aProgram.GetRelatedChildrenOfParentCollection();

      ////---------------------------------
      ////Navigation 
      ////---------------------------------
      //aProgram.NavigationBasic();
      //aProgram.NavigationBasicAsynchronous();
      //aProgram.NavigationBasicAsynchronousAnonymousCallback();
      //aProgram.NavigationSynchronousPreload();
      //aProgram.NavigationAsynchronousPreload();

      ////---------------------------------
      ////Asynchronous Communications with the BOS
      ////---------------------------------
      //aProgram.BOSCom_AsyncQuery();
      //aProgram.BOSCom_AsyncQueryLambda();

      ////---------------------------------
      ////EntityListManager 
      ////---------------------------------
      //aProgram.SetUpEntityListManager();
      //aProgram.SetUpEntityListManagerWithNamedDelegate();

      //aProgram.
      //aProgram.
      //aProgram.
      //aProgram.
      //aProgram.
      //aProgram.
      //aProgram.
      //aProgram.
      //aProgram.
    }

    #region Basic Queries

    private void BasicQuerySyntaxQuery() {
      ResetEntityManager(_em1);
      var customersQuery =
        from cust in _em1.Customers
        where cust.ContactTitle == "Sales Representative"
        orderby cust.CompanyName
        select cust;
      Console.WriteLine("Retrieved {0} customers", customersQuery.ToList().Count);
      PromptToContinue();
    }

    private void BasicMethodSyntaxQuery() {
      ResetEntityManager(_em1);
      var customersQuery = _em1.Customers
        .Where(c => c.ContactTitle == "Sales Representative")
        .OrderBy(c => c.CompanyName)
        .Select(c => c);
      Console.WriteLine("Retrieved {0} customers", customersQuery.ToList().Count);
      PromptToContinue();
    }

    private void MethodSyntaxShortForm() {
      ResetEntityManager(_em1);
      var customersQuery = _em1.Customers
        .Where(c => c.ContactTitle == "Sales Representative")
        .OrderBy(c => c.CompanyName);
      Console.WriteLine("Retrieved {0} customers", customersQuery.ToList().Count);
      PromptToContinue();
    }

    private void QueryWithExplicitlyTypedReturnValue() {
      ResetEntityManager(_em1);
      IEntityQuery<Customer> customersQuery = _em1.Customers
        .Where(c => c.ContactTitle == "Sales Representative")
        .OrderBy(c => c.CompanyName);
      Console.WriteLine("Retrieved {0} customers", customersQuery.ToList().Count);
      PromptToContinue();
    }

    private void RetrieveFirstCustomer() {
      ResetEntityManager(_em1);
      Customer firstCustomer = _em1.Customers
        .Where(c => c.ContactTitle == "Sales Representative")
        .OrderBy(c => c.CompanyName)
        .FirstOrNullEntity();
      Console.WriteLine("Retrieved {0} customers", _em1.Customers.With(QueryStrategy.CacheOnly).Count());
      PromptToContinue();
    }

    private void MixedQueryAndMethodSyntax() {
      ResetEntityManager(_em1);
      ICollection<Customer> customers =
      (from cust in _em1.Customers
       orderby cust.CompanyName
       select cust).ToList();
      DisplayCacheContents();
      PromptToContinue();
    }

    #endregion Basic Queries

    #region Forcing Execution

    private void ForceImmediateExecution() {
      ResetEntityManager(_em1);
      ICollection<Customer> customersQuery = _em1.Customers
        .Where(c => c.ContactTitle == "Sales Representative")
        .OrderBy(c => c.CompanyName)
        .ToList();
      Console.WriteLine("Retrieved {0} customers", customersQuery.Count());
      PromptToContinue();
    }

    private void ForceRetrievalUsingForEach() {
      ResetEntityManager(_em1);
      IEntityQuery<Customer> customersQuery = _em1.Customers
              .Where(c => c.ContactTitle == "Sales Representative")
              .OrderBy(c => c.CompanyName);

      foreach (Customer aCustomer in customersQuery) {
        // All customers are retrieved at the start of the loop
        Console.WriteLine("Retrieved {0} customers", _em1.Customers.With(QueryStrategy.CacheOnly).Count());
      }
      PromptToContinue();
    }

    #endregion Forcing Execution

    #region Skip and Take

    private void QueryWithSkipAndTake() {
      ResetEntityManager(_em1);
      IEntityQuery<Customer> customersQuery = _em1.Customers
              .Where(c => c.ContactTitle == "Sales Representative")
              .OrderBy(c => c.CompanyName)
              .With(QueryStrategy.DataSourceOnly);
      ICollection<Customer> customers = customersQuery.Skip(5).Take(5).ToList();
      Console.WriteLine("Retrieved {0} customers", _em1.Customers.With(QueryStrategy.CacheOnly).Count());
      PromptToContinue();
    }

    #endregion Skip and Take

    #region Queries with WITH

    private void QueriesWithWITH() {
      ResetEntityManager(_em1);

      IEntityQuery<Customer> query0 = _em1.Customers
      .Where(c => c.CompanyName.ToLower().StartsWith("a"));
      query0.QueryStrategy = QueryStrategy.DataSourceOnly;

      // Use With() to run the existing query against a different EntityManager:
      DomainModelEntityManager em2 = new DomainModelEntityManager();
      List<Customer> customers = new List<Customer>(query0.With(em2));

      // The next two examples use With() to run the query with a different QueryStrategy.

      // The With() call in the right-hand side of the following statement 
      // specifies a query that is materially different from query0, in
      // that it has a different QueryStrategy associated with it.
      // Accordingly, the right-hand side of the statement will return
      // a new query:
      IEntityQuery<Customer> query1 = query0.With(QueryStrategy.CacheOnly);

      // Because the content of the With() call in the right-hand side 
      // of the following statement doesn't result in a modification
      // of query0, the right-hand side will return a reference to 
      // query0 rather than a new query.
      IEntityQuery<Customer> query2 = query0.With(QueryStrategy.DataSourceOnly);

      // If you want to be certain you get a new query, use Clone() 
      // rather than With():
      EntityQuery<Customer> query3 = (EntityQuery<Customer>)query0.Clone();
      query3.QueryStrategy = QueryStrategy.DataSourceOnly;

      // Change both the QueryStrategy and the EntityManager
      IEntityQuery<Customer> query4 = query0.With(em2, QueryStrategy.CacheOnly);

      // You can pass null arguments to With(). When a query has a null EntityManager,
      // assigned, it uses the DefaultManager. When a query has a null QueryStrategy,
      // it uses the DefaultQueryStrategy of the assigned (or default) EntityManager.

      // Run the query against the default EntityManager, using its default QueryStrategy:
      IEntityQuery<Customer> query5 = query0.With(null, null);

      // When you pass a single null to With, you must cast it to the appropriate
      // type so the compiler know's which single-parameter overload you mean to use:

      // Run the query against the default EntityManager, using the base query's 
      // assigned QueryStrategy:
      IEntityQuery<Customer> query6 = query0.With((DomainModelEntityManager)null);

      // Run the query against the assigned EntityManager, using that EntityManager's
      // default QueryStrategy:
      IEntityQuery<Customer> query7 = query0.With((QueryStrategy)null);
    }

    #endregion Queries with WITH

    #region Using ToQuery()

    private void UsingToQueryPt01() {
      ResetEntityManager(_em1);
      Customer aCustomer = _em1.Customers.FirstOrNullEntity();
      var query = aCustomer.ToQuery<Customer>()
        .Include(Customer.PathFor(c => c.Orders));
      query.With(QueryStrategy.DataSourceOnly).ToList();
      DisplayCacheContents();
      PromptToContinue();
    }

    private void UsingToQueryPt02() {
      ResetEntityManager(_em1);
      List<Customer> customers = _em1.Customers
        .Where(c => c.CompanyName.ToLower().StartsWith("a")).ToList();
      var query2 = customers.ToQuery<Customer>();
      DisplayCacheContents();
      PromptToContinue();
    }

    #endregion Using ToQuery()

    #region ESQL and Stored Proc Queries

    private void PassThruEsqlQuery() {
      ResetEntityManager(_em1);
      PassthruEsqlQuery query = new PassthruEsqlQuery(typeof(Employee),
        "SELECT VALUE e FROM Employees AS e Where e.EmployeeID < 10");
      IEnumerable results = _em1.ExecuteQuery(query);
      DisplayCacheContents();
      PromptToContinue();
    }

    private void StoredProcQuery() {
      ResetEntityManager(_em1);
      QueryParameter param01 = new QueryParameter("EmployeeID", 1);
      QueryParameter param02 = new QueryParameter("Year", 1996);
      StoredProcQuery query = new StoredProcQuery(typeof(Order));
      query.Parameters.Add(param01);
      query.Parameters.Add(param02);

      // Note that a FunctionImport must be defined in the Entity Model
      query.ProcedureName = "OrdersGetForEmployeeAndYear";
      _em1.ExecuteQuery(query);
      DisplayCacheContents();
      PromptToContinue();
    }

    #endregion ESQL and Stored Proc Queries

    #region Using the PredicateBuilder

    private void ProductsWithNamesThatContainSpecifiedString() {
      ResetEntityManager(_em1);
      var word = "Sir";
      var q = _em1.Products
        .Where(p => p.ProductName.Contains(word));
      var results = q.ToList();// returns 3 Northwind products
      DisplayCacheContents();
      PromptToContinue();
    }

    private void ProductsWithNamesThatContainSpecifiedStrings() {
      ResetEntityManager(_em1);
      var q = _em1.Products
        .Where(p =>
          p.ProductName.Contains("Sir") ||
          p.ProductName.Contains("Cajun") ||
          p.ProductName.Contains("Louisiana")
          );
      var results = q.ToList(); // returns 6 Northwind products
      DisplayCacheContents();
      PromptToContinue();
    }

    private void PredicateBuilder01() {
      ResetEntityManager(_em1);
      var words = GetWords("Sir Cajun Louisiana");
      var tests = ProductNameTests(words).ToArray();
      if (0 == tests.Length) return;
      var productNamePredicate = PredicateBuilder.Or(tests);
      var q = _em1.Products.Where(productNamePredicate);
      var results = q.ToList(); // returns 6 Northwind products

      DisplayCacheContents();
      PromptToContinue();
    }

    /// <summary>
    /// A quick tutorial on building predicate expressions with the PredicateBuilder
    /// </summary>
    /// <remarks>
    /// Set breakpoint on first line, step through, and watch
    /// the value of variables in the Autos window
    /// </remarks>
    private void PredicateBuilderMiscExamples() {
      ResetEntityManager(_em1);
      var words = GetWords("Sir Cajun Louisiana");
      var tests = ProductNameTests(words).ToArray();

      Expression<Func<Product, bool>> p1, p2, p3, p4, bigP;

      // Sample predicate expressions
      p1 = p => p.ProductName.Contains("Sir");
      p2 = p => p.ProductName.Contains("Cajun");
      p3 = p => p.ProductName.Contains("Louisiana");
      p4 = p => p.UnitPrice > 20;

      bigP = p1.Or(p2); // Name contains "Sir" or "Cajun"

      bigP = p1.Or(p2).Or(p3); // Name contains any of the three

      bigP = PredicateBuilder.Or(p1, p2, p3); // Name contains any of the 3

      bigP = PredicateBuilder.Or(tests); // OR together some tests

      bigP = p1.And(p4); // "Sir" and price is greater than 20

      // Name contains "Cajun" and "Lousiana" and the price is greater than 20
      bigP = PredicateBuilder.And(p2, p3, p4);

      bigP = PredicateBuilder.And(tests); // AND together some tests

      // Name contains either “Sir” or “Louisiana” AND price is greater than 20 
      bigP = p1.Or(p3).And(p4); // 

      bigP = PredicateBuilder.Not(p1); // Name does not contain "Sir"

      bigP = PredicateBuilder.True<Product>().And(p1);// same as p1
      bigP = PredicateBuilder.False<Product>().Or(p1);// same as p1

      // Not useful
      bigP = PredicateBuilder.True<Product>().Or(p1);// always true
      bigP = PredicateBuilder.False<Product>().And(p1);// always false
    }

    private void PredicateDescriptions01() {
      ResetEntityManager(_em1);

      PredicateDescription p1 = PredicateBuilder.Make(typeof(Product), "UnitPrice",
        FilterOperator.IsGreaterThanOrEqualTo, 24);
      PredicateDescription p2 = PredicateBuilder.Make(typeof(Product), "Discontinued",
        FilterOperator.IsEqualTo, true);

      var query = PredicateBuilder.FilterQuery(_em1.Products, p1.And(p2));
      var results = _em1.ExecuteQuery<Product>((IEntityQuery<Product>)query);
      // The above query is the same as: 
      //var queryb = _entityManager.Products.Where(p => p.UnitPrice > 24 && p.Discontinued);
      DisplayCacheContents();
      PromptToContinue();
    }

    private void CompositePredicateDescription() {
      ResetEntityManager(_em1);
      PredicateDescription p1 = new PredicateDescription(typeof(Product),
        "UnitPrice", FilterOperator.IsGreaterThanOrEqualTo, 24);
      PredicateDescription p2 = new PredicateDescription(typeof(Product),
        "Discontinued", FilterOperator.IsEqualTo, true);

      // And the two filters.
      CompositePredicateDescription p3 = p1.And(p2);
    }

    private void CompositePredicateDescriptionToLambda() {
      ResetEntityManager(_em1);
      PredicateDescription p1 = new PredicateDescription(typeof(Product),
        "UnitPrice", FilterOperator.IsGreaterThanOrEqualTo, 24);
      PredicateDescription p2 = new PredicateDescription(typeof(Product),
        "Discontinued", FilterOperator.IsEqualTo, true);

      // And the two filters.
      CompositePredicateDescription p3 = p1.And(p2);

      var exprFunc = (Expression<Func<Product, bool>>)p3.ToLambdaExpression();
      var filterQuery = _em1.Products.Where(exprFunc);
      var results = _em1.ExecuteQuery(filterQuery);

      DisplayCacheContents();
      PromptToContinue();
    }

    /// <summary>
    /// Given an arbitrary collection of entities of a particular type,
    /// build a query that will retrieve all of their related children
    /// of another type.
    /// 
    /// Here we find related Orders for a collection of Customers.
    /// </summary>
    private void GetRelatedChildrenOfParentCollection() {
      ResetEntityManager(_em1);

      // Start with a list of customers that you’ve populated however you see fit,
      // perhaps from end-user input. Here, we’ll arbitrarily populate one as follows:
      List<Customer> customers = _em1.Customers.Where(c => c.Country == "USA").ToList();
      customers.AddRange(_em1.Customers.Where(c => c.Country == "Brazil").ToList());

      // From the list of customers, create an IEnumerable<PredicateDescription>
      var predicates = customers.Select(c => new PredicateDescription(typeof(Customer),
        "CustomerId", FilterOperator.IsEqualTo, c.CustomerID));

      // Convert that IEnumerable<PredicateDescription> to an array, and feed the array
      // to the PredicateBuilder’s Or() method to get a CompositePredicateDescription
      var customerPredicate = PredicateBuilder.Or(predicates.ToArray());

      // Convert the CompositePredicateDescription to a LambdaExpression; pass the
      // LambdaExpression to the Where clause of a Customer query; and project out
      // the related Orders using a SelectMany() call. Execute the query to retrieve
      // the desired Orders!
      var exprFunc = (Expression<Func<Customer, bool>>)customerPredicate.ToLambdaExpression();
      var ordersQuery = _em1.Customers.Where(exprFunc).SelectMany(c => c.Orders);
      var orders = _em1.ExecuteQuery(ordersQuery);

      DisplayCacheContents();
      PromptToContinue();
    }

    #endregion Using the PredicateBuilder

    #region PredicateBuilder Auxiliary Methods

    private IEnumerable<String> GetWords(string phrase) {
      return phrase.Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
    }

    private IEnumerable<Expression<Func<Product, bool>>>
        ProductNameTests(IEnumerable<String> words) {
      foreach (var each in words) {
        var word = each; // must include this statement so that each is evaluated at each iteration
        yield return p => p.ProductName.Contains(word);
      }
    }

    /// <summary>
    /// This version of ProductNameTests contains no yield statement, so that it
    /// can be automatically translated to VB, which does not support yield.
    /// </summary>
    /// <param name="words"></param>
    /// <returns></returns>
    private IEnumerable<Expression<Func<Product, bool>>>
        ProductNameTestsForTranslationToVB(IEnumerable<String> words) {
      List<Expression<Func<Product, bool>>> expressions = new List<Expression<Func<Product, bool>>>();
      foreach (var each in words) {
        var word = each; // must include this statement so that each is evaluated at each iteration
        expressions.Add(p => p.ProductName.Contains(word));
      }
      return expressions;
    }

    #endregion PredicateBuilder Auxiliary Methods

    #region PassthruESQL Queries
    public void EsqlBasic() {
      var query = new PassthruEsqlQuery(typeof(Customer),
        "SELECT VALUE c FROM Customers AS c Where c.Country == 'Brazil'");
      var result = query.With(_em1).Execute().Cast<Customer>();
      Console.WriteLine("Retrieved {0} Customers", result.Count());
      PromptToContinue();
    }

    public void EsqlWithParameter() {
      var param = new QueryParameter("country", "Brazil");
      var paramEsql = new ParameterizedEsql(
        "SELECT VALUE c FROM Customers AS c Where c.Country > @country", param);
      var query = new PassthruEsqlQuery(typeof(Customer), paramEsql);
      var result1 = query.With(_em1).Execute().Cast<Customer>();
      Console.WriteLine("Retrieved {0} Customers from {1}", result1.Count(), param.Value.ToString());
      //Assert.IsTrue(r1.Count() > 5);
      param.Value = "Germany";
      var result2 = query.With(_em1).Execute().Cast<Customer>();
      Console.WriteLine("Retrieved {0} Customers from {1}", result2.Count(), param.Value.ToString());
      PromptToContinue();
    }

    #endregion PassthruESQL Queries

    #region Navigation

    #region NavigationBasic

    public void NavigationBasic() {
      Order anOrder = _em1.Orders.FirstOrNullEntity();
      if (!anOrder.EntityAspect.IsNullEntity) {
        List<OrderDetail> lineItems = new List<OrderDetail>(anOrder.OrderDetails);
        Console.WriteLine("lineItems.Count = {0}", lineItems.Count);
        OrderDetail firstItem = lineItems[0];
        firstItem.Quantity = 10;
        Console.WriteLine("lineItem[0].Quantity = {0}", lineItems[0].Quantity);
      }
      PromptToContinue();
    }
    #endregion NavigationBasic

    #region NavigationBasicAsynchronous
    public void NavigationBasicAsynchronous() {
      _em1.UseAsyncNavigation = true;
      IEntityQuery<Order> query = _em1.Orders.Where(o => o.OrderID == 10248);
      _em1.ExecuteQueryAsync<Order>(query, GotOrder, null);
      PromptToContinue();
    }

    private void GotOrder(EntityQueryOperation<Order> op) {
      if (op.Error != null) {
        Console.WriteLine(op.Error.Message);
      }
      else {
        // Retrieve a single related entity using a scalar navigation property
        Order targetOrder = op.Results.ToList()[0];
        Console.WriteLine("Order: {0}", targetOrder.OrderID.ToString());
        targetOrder.Customer.EntityAspect.PendingEntityResolved +=
          new EventHandler<PendingEntityResolvedEventArgs>(
            Customer_PendingEntityResolved);
        Customer aCustomer = targetOrder.Customer;
        Console.WriteLine("Customer (from GotOrders): {0}",
          aCustomer.CompanyName);

        // Retrieve a collection of related entities using a collection navigation property
        targetOrder.OrderDetails.PendingEntityListResolved +=
          new EventHandler<PendingEntityListResolvedEventArgs<OrderDetail>>(
            OrderDetails_PendingEntityListResolved);
      }
    }

    void Customer_PendingEntityResolved(object sender,
        PendingEntityResolvedEventArgs e) {
      Customer customer = (Customer)e.ResolvedEntity;
      Console.WriteLine("Customer (from Customer_PendingEntityResolved): {0}",
        customer.CompanyName);
    }

    void OrderDetails_PendingEntityListResolved(object sender,
        PendingEntityListResolvedEventArgs<OrderDetail> e) {
      Console.WriteLine("OrderDetails retrieved: {0}", e.ResolvedEntities.Count);
    }

    #endregion NavigationBasicAsynchronous

    #region NavigationBasicAsynchronousAnonymousCallback
    public void NavigationBasicAsynchronousAnonymousCallback() {
      _em1.UseAsyncNavigation = true;
      IEntityQuery<Order> query = _em1.Orders.Where(o => o.OrderID == 10248);
      _em1.ExecuteQueryAsync<Order>(
        query,                                           // IEntityQuery<Order>
        (args) => {                                      // AsyncCompletedCallback
          Console.WriteLine("Order: {0}",                //    "
            ((Order)args.Results.ToList()[0]).OrderID);   //    "
        },                                               //    "
      null                                               // UserState object
      );

      PromptToContinue();
    }
    #endregion NavigationBasicAsynchronousAnonymousCallback

    #region Navigation - Preload Using a Span Query
    public void NavigationSynchronousPreload() {
      ResetEntityManager(_em1);
      IEntityQuery<Customer> query = _em1.Customers.Where(c => c.Country == "France")
        .Include("Orders")
        .Include("Orders.OrderDetails")
        .Include("Orders.OrderDetails.Product")
        .Include("Orders.OrderDetails.Product.Supplier")
        .Include("Orders.SalesRep");
      _em1.ExecuteQuery<Customer>(query);
      // _em1.ExecuteQuery(query); // accomplishes the same thing
      // query.ToList();           // accomplishes the same thing
      PromptToContinue();
    }

    public void NavigationAsynchronousPreload() {
      ResetEntityManager(_em1);
      _em1.UseAsyncNavigation = true;
      IEntityQuery<Customer> query = _em1.Customers.Where(c => c.Country == "France")
        .Include("Orders")
        .Include("Orders.OrderDetails")
        .Include("Orders.OrderDetails.Product")
        .Include("Orders.OrderDetails.Product.Supplier")
        .Include("Orders.SalesRep");
      _em1.ExecuteQueryAsync<Customer>(query, GotCustomers, null);
      PromptToContinue();
    }

    private void GotCustomers(EntityQueryOperation<Customer> op) {
      if (op.Error != null) {
        Console.WriteLine(op.Error.Message);
      }
      else {
        DisplayCacheContents();
      }
    }

    #endregion Navigation - Asynchronous Preload Pattern

    #endregion Navigation

    #region Asynchronous Communications with the BOS

    private void BOSCom_AsyncQuery() {
      ResetEntityManager(_em1);
      var query = new EntityQuery<Customer>()
        .Where(c => c.Country == "Denmark");
      int token = 1;
      _em1.ExecuteQueryAsync<Customer>(query,
        QueryCompletedCallback, token);
    }

    private void QueryCompletedCallback(EntityQueryOperation<Customer> args) {
      var resultList = args.Results;
      Console.WriteLine("Query returned {0} entities", resultList.Count());
      PromptToContinue();
    }

    private void BOSCom_AsyncQueryLambda() {
      var query = new EntityQuery<Customer>().Where(c => c.Country == "Denmark");
      int token = 2;
      _em1.ExecuteQueryAsync<Customer>(
          query,
          (args) => {
            var resultList = args.Results;
            Console.WriteLine("Query returned {0} entities", resultList.Count());
          },
        token);
      PromptToContinue();
    }

    #endregion Asynchronous Communications with the BOS

    #region EntityListManager

    private void SetUpEntityListManager() {
      ResetEntityManager(_em1);
      _em1.Employees.ToList(); // load Employees into cache
      _salesReps.Clear();
      var filter = new Predicate<Employee>(
          delegate(Employee anEmployee) { return anEmployee.City == "London"; });
      _employeeEntityListManager =
          new EntityListManager<Employee>(_em1, filter, null);

      // The following is more efficient, since it causes the list manager
      // only to react only to changes in the Employee.City property,
      // and to ignore all other changes.

      //_employeeEntityListManager =
      //    new EntityListManager<Employee>(_em1, filter,
      //      new EntityProperty[] { Employee.CityEntityProperty });

      bool refreshListWhenPlacedUnderManagement = true;
      _employeeEntityListManager.ManageList(_salesReps,
        refreshListWhenPlacedUnderManagement);

      foreach (Employee anEmployee in _salesReps) {
        Console.WriteLine(anEmployee.LastName);
      }
      PromptToContinue();
    }

    private void SetUpEntityListManagerWithNamedDelegate() {
      // Identify Customer currently being edited by some process;
      // this is a stand-in.
      _currentCustomer = _em1.Customers.FirstOrNullEntity();

      EntityListManager<Order> orderEntityListManager =
        new EntityListManager<Order>(_em1, FilterOrdersByDate,
        new EntityProperty[] {
          Order.PropertyMetadata.OrderDate, 
          Order.PropertyMetadata.Customer  }
        );
    }

    /// <summary>
    /// This rule gets the 1996 Orders for the current Customer
    /// </summary>
    /// <param name="pOrder"></param>
    /// <returns></returns>
    Boolean FilterOrdersByDate(Order pOrder) {
      return (pOrder.OrderDate.Value.Year == 1996 &&
        pOrder.Customer == _currentCustomer);
    }

    #endregion EntityListManager

    #region QueryStrategy

    private void QueryStrategyAssortedSyntaxExamples() {

      // Setting the QueryStrategy
      EntityQuery<Order> query01 = _em1.Orders;
      query01.QueryStrategy = QueryStrategy.DataSourceThenCache;

      // Setting the EntityManager's DefaultQueryStrategy
      _em1.DefaultQueryStrategy = QueryStrategy.Normal;

      // Query that returns a scalar result
      var query02 = _em1.Orders.Select(o => o.Freight).Sum();

      // Query that returns a single element
      var query03 = _em1.Products.OrderByDescending(c => c.ProductName).FirstOrNullEntity();

      // Query whose return type is different from the type contained 
      // in the collection first referenced.
      var query04 = _em1.Customers
        .Where(c => c.Country == "Argentina")
        .SelectMany(c => c.Orders);

      // Assigning a pre-defined QueryStrategy
      query04.QueryStrategy = QueryStrategy.DataSourceThenCache;

      // Creating a custom QueryStrategy
      QueryStrategy aQueryStrategy =
        new QueryStrategy(FetchStrategy.DataSourceThenCache,
        MergeStrategy.PreserveChanges,
        QueryInversionMode.On);
    }

    private void QueryStrategyWithAndCloning() {
      IEntityQuery<Customer> query00 = _em1.Customers
        .Where(c => c.CompanyName.ToLower().StartsWith("a"));
      query00.QueryStrategy = QueryStrategy.DataSourceOnly;

      // The With() call in the right-hand side of the following statement 
      // specifies a query that is materially different from query0, in
      // that it has a different QueryStrategy associated with it.
      // Accordingly, the right-hand side of the statement will return
      // a new query:
      IEntityQuery<Customer> query01 = query00.With(QueryStrategy.CacheOnly);

      // Because the content of the With() call in the right-hand side 
      // of the following statement doesn't result in a modification
      // of query0, the right-hand side will return a reference to 
      // query0 rather than a new query.
      IEntityQuery<Customer> query02 = query00.With(QueryStrategy.DataSourceOnly);

      // If you want to be certain you get a new query, use Clone() 
      // rather than With():
      EntityQuery<Customer> query03 = (EntityQuery<Customer>)query00.Clone();
      query03.QueryStrategy = QueryStrategy.DataSourceOnly;
    }

    #endregion QueryStrategy

    #region General-Purpose Methods

    private void ResetEntityManager(EntityManager em) {
      em.Clear();
      em.UseAsyncNavigation = false;
    }

    private void DisplayCacheContents() {
      Console.WriteLine("Contents of Cache");
      Console.WriteLine("-----------------");
      Console.WriteLine("Customers: {0}", _em1.Customers.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Employees: {0}", _em1.Employees.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Orders: {0}", _em1.Orders.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("OrderDetails: {0}", _em1.OrderDetails.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Products: {0}", _em1.Products.With(QueryStrategy.CacheOnly).Count());
      Console.WriteLine("Suppliers: {0}", _em1.Suppliers.With(QueryStrategy.CacheOnly).Count());
    }

    private void PromptToContinue() {
      Console.WriteLine();
      Console.WriteLine("Press ENTER to continue...");
      Console.ReadLine();
    }

    #endregion General-Purpose Methods

    #region Private Fields
    DomainModelEntityManager _em1 = new DomainModelEntityManager();
    EntityListManager<Employee> _employeeEntityListManager;
    List<Employee> _salesReps = new List<Employee>();
    Customer _currentCustomer;
    #endregion Private Fields
  }
}
